import logging
from typing import Tuple, NamedTuple, TypeVar

from etils.array_types import PRNGKey, IntArray, BoolArray
from jax import tree_map, numpy as jnp, random, pmap
from jax._src.lax.control_flow import scan, while_loop
from jax._src.lax.parallel import all_gather

from jaxns.model import Model
from jaxns.slice_sampler import PreprocessType, AbstractSliceSampler
from jaxns.statistics import analyse_sample_collection
from jaxns.termination import determine_termination
from jaxns.types import NestedSamplerState, Reservoir, LivePoints, TerminationCondition, int_type
from jaxns.utils import collect_samples

logger = logging.getLogger('jaxns')

__all__ = ['StaticSlice']

T = TypeVar('T')


def remove_chunk_dim(py_tree: T) -> T:
    def _remove_chunk_dim(a):
        shape = list(a.shape)
        shape = [shape[0] * shape[1]] + shape[2:]
        return jnp.reshape(a, shape)

    return tree_map(_remove_chunk_dim, py_tree)


def add_chunk_dim(py_tree: T, chunk_size: int) -> T:
    def _add_chunk_dim(a):
        shape = list(a.shape)
        shape = [chunk_size, shape[0] // chunk_size] + shape[1:]
        return jnp.reshape(a, shape)

    return tree_map(_add_chunk_dim, py_tree)


class StaticSlice:
    """
    Performs parallel nested sampling with a static number of samples.
    This uses the fact that the union of samples from N independent perfect static nested samplers with M live points is
    equivalent to a single nested sampler with N*M live points.
    """

    def __init__(self, model: Model, slice_sampler: AbstractSliceSampler, num_live_points: int, num_slices: int,
                 num_parallel_samplers: int = 1):
        if num_live_points % num_parallel_samplers != 0:
            raise ValueError(
                f"num_live_points {num_live_points} must divide num_parallel_samplers {num_parallel_samplers}"
            )
        self.num_live_points = num_live_points
        self.num_parallel_samplers = num_parallel_samplers
        self.model = model
        self.slice_sampler = slice_sampler
        if num_slices < 1:
            raise ValueError(f"num_slices should be > 0, got {num_slices}.")
        self.num_slices = num_slices

    def _single_live_point_shrink(self,
                                  key: PRNGKey,
                                  live_points: LivePoints,
                                  num_slices: IntArray,
                                  preprocess_data: PreprocessType) -> Tuple[Reservoir, LivePoints]:
        """
        Run nested sampling to replace an entire live point reservoir via shrinkage.

        Args:
            key: PRNGKey
            live_points: live points
            num_slices: slice sampling `num_slices` parameter

        Returns:
            dead point reservoir, live points, slice stats
        """

        class CarryType(NamedTuple):
            key: PRNGKey
            live_points: LivePoints

        ResultType = Reservoir

        def body(carry: CarryType, unused_X: IntArray) -> Tuple[CarryType, ResultType]:
            del unused_X
            idx_min = jnp.argmin(carry.live_points.reservoir.log_L)
            dead_point: Reservoir = tree_map(lambda x: x[idx_min], carry.live_points.reservoir)
            log_L_dead = dead_point.log_L

            # contour becomes log_L_dead if log_L_dead is not supremum of live-points, else we choose the original
            # constraint of dead point. Note: we are at liberty to choose any log_L level as a contour so long as we
            # can sample within it uniformly.
            on_supremum = jnp.equal(log_L_dead, jnp.max(carry.live_points.reservoir.log_L))
            log_L_contour = jnp.where(on_supremum, dead_point.log_L_constraint, log_L_dead)

            # replace dead point with a new sample about contour
            key, seed_key, sample_key = random.split(carry.key, 3)

            seed_point = self.slice_sampler.get_seed_point(key=seed_key,
                                                           live_points=carry.live_points,
                                                           log_L_constraint=log_L_contour)
            sample = self.slice_sampler.get_sample(key=sample_key,
                                                   seed_point=seed_point,
                                                   log_L_constraint=log_L_contour,
                                                   num_slices=num_slices,
                                                   preprocess_data=preprocess_data)

            live_points_reservoir = tree_map(lambda old, update: old.at[idx_min].set(update),
                                             carry.live_points.reservoir, Reservoir(*sample))
            live_points = carry.live_points._replace(reservoir=live_points_reservoir)

            return CarryType(key, live_points), dead_point

        (_, live_points), dead_reservoir = scan(body,
                                                CarryType(key, live_points),
                                                live_points.reservoir.log_L)
        return (dead_reservoir, live_points)

    def _single_thread_ns(self, state: NestedSamplerState, live_points: LivePoints,
                          termination_cond: TerminationCondition):

        class CarryType(NamedTuple):
            done: BoolArray
            termination_reason: IntArray
            state: NestedSamplerState
            live_points: LivePoints

        def body(body_state: CarryType) -> CarryType:
            key, sample_key = random.split(body_state.state.key, 2)
            state = body_state.state._replace(key=key)

            preprocess_data = self.slice_sampler.preprocess(state)

            (dead_reservoir, live_points) = self._single_live_point_shrink(
                key=sample_key,
                live_points=body_state.live_points,
                num_slices=self.num_slices,
                preprocess_data=preprocess_data
            )

            all_dead_reservoir: Reservoir = remove_chunk_dim(all_gather(dead_reservoir, 'i'))

            # update the state with sampled points from last round
            new_state = collect_samples(state, all_dead_reservoir)

            evidence_calculation, sample_stats = analyse_sample_collection(
                sample_collection=new_state.sample_collection,
                sorted_collection=True
            )
            done, termination_reason = determine_termination(
                term_cond=termination_cond,
                sample_collection=new_state.sample_collection,
                evidence_calculation=evidence_calculation,
                live_points=live_points
            )

            return CarryType(done=done, termination_reason=termination_reason, state=new_state,
                             live_points=live_points)

        output_carry = while_loop(
            lambda body_state: jnp.bitwise_not(body_state[0]),
            body,
            CarryType(done=jnp.asarray(False, jnp.bool_),
                      termination_reason=jnp.asarray(0, int_type), state=state, live_points=live_points)
        )

        return output_carry.termination_reason, output_carry.state, output_carry.live_points

    def __call__(self, state: NestedSamplerState,
                 live_points: LivePoints,
                 termination_cond: TerminationCondition
                 ) -> Tuple[IntArray, NestedSamplerState, LivePoints]:
        """
        Performs nested sampling from an initial state with a static number of live points at each shrinkage step, and
        terminating upon ANY of the terminiation criteria being met.

        Args:
            state: the current state of sampler
            live_points: the live points to start sampling from
            termination_cond: the conditions for termination

        Returns:
            termination_reason, final state, and live points
        """
        if live_points.reservoir.log_L.size != self.num_live_points:
            raise ValueError(
                f"live points reservoir is the wrong size. "
                f"Got {live_points.reservoir.log_L.size} but expected {self.num_live_points}.")

        chunked_live_points = add_chunk_dim(live_points, self.num_parallel_samplers)

        parallel_ns = pmap(lambda live_points: self._single_thread_ns(
            state=state,
            live_points=live_points,
            termination_cond=termination_cond
        ), axis_name='i')

        chunked_termination_reason, chunked_state, chunked_live_points = parallel_ns(chunked_live_points)

        termination_reason, state = tree_map(lambda x: x[0], (chunked_termination_reason, chunked_state))
        live_points = remove_chunk_dim(chunked_live_points)

        return termination_reason, state, live_points



import io
import logging
from typing import NamedTuple, TextIO, Union, Optional

import numpy as np
from jax import numpy as jnp, tree_map, vmap, random, jit
from jax._src.lax.control_flow import while_loop

from jaxns.internals.log_semiring import LogSpace
from jaxns.internals.maps import replace_index, prepare_func_args
from jaxns.model import Model
from jaxns.random import resample_indicies
from jaxns.types import SampleCollection, NestedSamplerState, Reservoir, NestedSamplerResults, \
    float_type

logger = logging.getLogger('jaxns')

__all__ = ['sort_samples',
           'collect_samples',
           'resample',
           'marginalise_static',
           'marginalise_dynamic',
           'maximum_a_posteriori_point',
           'summary',
           'evidence_posterior_samples',
           'analytic_posterior_samples',
           'analytic_log_evidence',
           'save_pytree',
           'save_results',
           'load_pytree',
           'load_results']

def sort_samples(sample_collection: SampleCollection):
    """
    Sorts a sample collection lexigraphically, first on log_L then on log_L_constraint.

    Args:
        sample_collection: sample collection to sort

    Returns:
        sorted sample collection
    """
    idx_sort = jnp.lexsort((sample_collection.reservoir.log_L_constraint,
                            sample_collection.reservoir.log_L))
    sample_collection = sample_collection._replace(
        reservoir=tree_map(lambda x: x[idx_sort], sample_collection.reservoir)
    )
    return sample_collection


def collect_samples(state: NestedSamplerState, new_reservoir: Reservoir) -> NestedSamplerState:
    """
    Merge samples from new reservoir into sample collection

    Args:
        state: state
        new_reservoir: reservoir to add to sample collection

    Returns:
        state with sorted sample collection
    """
    old_reservoir = state.sample_collection.reservoir
    # Insert the new samples with a slice update
    reservoir = tree_map(lambda old, new: replace_index(old, new, state.sample_collection.sample_idx),
                         old_reservoir, new_reservoir)
    # Update the sample collection
    sample_collection = state.sample_collection._replace(
        reservoir=reservoir,
        sample_idx=state.sample_collection.sample_idx + new_reservoir.log_L.size
    )
    # Sort
    sample_collection = sort_samples(sample_collection)
    state = state._replace(sample_collection=sample_collection)
    return state


def resample(key, samples, log_weights, S=None, replace=False):
    idx = resample_indicies(key, log_weights, S=S, replace=replace)
    return tree_map(lambda s: s[idx, ...], samples)


def marginalise_static(key, samples, log_weights, ESS, fun):
    """
    Marginalises function over posterior samples, where ESS is static.

    Args:
        key: PRNG key
        samples (dict): dict of batched array of nested sampling samples
        log_weights: log weights from nested sampling
        ESS: static effective sample size
        fun (:code:`callable(**kwargs)`): function to marginalise

    Returns:
        expectation over resampled samples.
    """
    fun = prepare_func_args(fun)
    samples = resample(key, samples, log_weights, S=ESS, replace=True)
    marginalised = tree_map(lambda marg: jnp.nanmean(marg, axis=0), vmap(fun)(**samples))
    return marginalised


def marginalise_dynamic(key, samples, log_weights, ESS, fun):
    """
    Marginalises function over posterior samples, where ESS can be dynamic.

    Args:
        key: PRNG key
        samples (dict): dict of batched array of nested sampling samples
        log_weights: log weights from nested sampling
        ESS: dynamic effective sample size
        fun (:code:`callable(**kwargs)`): function to marginalise

    Returns:
        expectation over resampled samples.
    """
    fun = prepare_func_args(fun)
    ESS = jnp.asarray(ESS)

    def body(state):
        (key, i, count, marginalised) = state
        key, resample_key = random.split(key, 2)
        _samples = resample(resample_key, samples, log_weights, S=1)
        _sample = tree_map(lambda v: v[0], _samples)
        update = fun(**_sample)
        count = tree_map(lambda y, c: jnp.where(jnp.any(jnp.isnan(y)), c, c + jnp.asarray(1, c.dtype)),
                         update, count)
        marginalised = tree_map(lambda x, y: jnp.where(jnp.isnan(y), x, x + y.astype(x.dtype)),
                                marginalised, update)
        return (key, i + jnp.ones_like(i), count, marginalised)

    test_output = fun(**tree_map(lambda v: v[0], samples))
    count = tree_map(lambda x: jnp.asarray(0, x.dtype), test_output)
    init_marginalised = tree_map(lambda x: jnp.zeros_like(x), test_output)
    (_, _, count, marginalised) = while_loop(lambda state: state[1] < ESS,
                                             body,
                                             (key, jnp.array(0, ESS.dtype), count, init_marginalised))
    marginalised = tree_map(lambda x, c: x / c, marginalised, count)
    return marginalised


def maximum_a_posteriori_point(results: NestedSamplerResults):
    """
    Get the MAP point of a nested sampling result.
    Does this by choosing the point with largest L(x) p(x).

    Args:
        results (NestedSamplerResult): Nested sampler result

    Returns:
        dict of samples at MAP-point.
    """

    map_idx = jnp.argmax(results.log_posterior_density)
    map_points = tree_map(lambda x: x[map_idx], results.samples)
    return map_points


def _bit_mask(int_mask, width=8):
    """
    Convert an integer mask into a bit-mask. I.e. convert an integer into list of left-starting bits.

    Examples:

    1 -> [1,0,0,0,0,0,0,0]
    2 -> [0,1,0,0,0,0,0,0]
    3 -> [1,1,0,0,0,0,0,0]

    Args:
        int_mask: int
        width: number of output bits

    Returns:
        List of bits from left
    """
    return list(map(int, '{:0{size}b}'.format(int_mask, size=width)))[::-1]


def summary(results: NestedSamplerResults, f_obj: Optional[Union[str, TextIO]] = None):
    """
    Gives a summary of the results of a nested sampling run.

    Args:
        results (NestedSamplerResults): Nested sampler result
    """
    main_s = []

    def _print(s):
        print(s)
        main_s.append(s)

    def _round(v, uncert_v):
        try:
            sig_figs = -int("{:e}".format(uncert_v).split('e')[1]) + 1
            return round(float(v), sig_figs)
        except:
            return float(v)

    _print("--------")
    termination_bit_mask = _bit_mask(results.termination_reason, width=6)
    _print("Termination Conditions:")
    for bit, condition in zip(termination_bit_mask, ['Reached max samples',
                                                     'Evidence uncertainty low enough',
                                                     'Small remaining evidence',
                                                     'Reached ESS',
                                                     "Used max num likelihood evaluations",
                                                     'likelihood contour reached']):
        if bit == 1:
            _print(condition)
    _print("--------")
    _print("# likelihood evals: {}".format(results.total_num_likelihood_evaluations))
    _print("# samples: {}".format(results.total_num_samples))
    _print("# slices: {:.1f}".format(results.total_num_slices))
    _print(
        "# slices / acceptance: {:.1f}".format(
            jnp.nanmean(jnp.where(results.num_slices > 0, results.num_slices, jnp.nan))))
    _print("# likelihood evals / sample: {:.1f}".format(
        results.total_num_likelihood_evaluations / results.total_num_samples))
    likelihood_evals_per_slice = jnp.sum(
        jnp.where(results.num_slices > 0, results.num_likelihood_evaluations_per_sample, 0.)) / results.total_num_slices
    _print("# likelihood evals / slice: {:.1f}".format(likelihood_evals_per_slice))
    _print("--------")
    _print("logZ={} +- {}".format(_round(results.log_Z_mean, results.log_Z_uncert),
                                  _round(results.log_Z_uncert, results.log_Z_uncert)))
    # _print("H={} +- {}".format(
    #     _round(results.H_mean, results.H_uncert), _round(results.H_uncert, results.H_uncert)))
    _print("H={}".format(
        _round(results.H_mean, results.H_mean)))
    _print("ESS={}".format(int(results.ESS)))
    max_like_idx = jnp.argmax(results.log_L_samples)
    max_like_points = tree_map(lambda x: x[max_like_idx], results.samples)
    samples = resample(random.PRNGKey(23426), results.samples, results.log_dp_mean, S=max(10, int(results.ESS)),
                       replace=True)

    max_map_idx = jnp.argmax(results.log_posterior_density)
    map_points = tree_map(lambda x: x[max_map_idx], results.samples)

    for name in samples.keys():
        _samples = samples[name].reshape((samples[name].shape[0], -1))
        _max_like_points = max_like_points[name].reshape((-1,))
        _map_points = map_points[name].reshape((-1,))
        ndims = _samples.shape[1]
        _print("--------")
        _print("{}: mean +- std.dev. | 10%ile / 50%ile / 90%ile | MAP est. | max(L) est.".format(
            name if ndims == 1 else "{}[#]".format(name), ))
        for dim in range(ndims):
            _uncert = jnp.std(_samples[:, dim])
            _max_like_point = _max_like_points[dim]
            _map_point = _map_points[dim]
            # two sig-figs based on uncert
            sig_figs = -int("{:e}".format(_uncert).split('e')[1]) + 1

            def _round(ar):
                return round(float(ar), sig_figs)

            _uncert = _round(_uncert)
            _print("{}: {} +- {} | {} / {} / {} | {} | {}".format(
                name if ndims == 1 else "{}[{}]".format(name, dim),
                _round(jnp.mean(_samples[:, dim])), _uncert,
                *[_round(a) for a in jnp.percentile(_samples[:, dim], jnp.asarray([10, 50, 90]))],
                _round(_map_point),
                _round(_max_like_point)
            ))
    _print("--------")
    if f_obj is not None:
        out = "\n".join(main_s)
        if isinstance(f_obj, str):
            with open(f_obj, 'w') as f:
                f.write(out)
        elif isinstance(f_obj, io.TextIOBase):
            f_obj.write(out)
        else:
            raise TypeError(f"Invalid f_obj: {type(f_obj)}")


def evidence_posterior_samples(key, num_live_points_per_sample, log_L_samples, S: int = 100):
    n_i = num_live_points_per_sample
    L = LogSpace(jnp.asarray([-jnp.inf], log_L_samples.dtype)).concatenate(
        LogSpace(log_L_samples))
    L_mid = (L[:-1] + L[1:]) * LogSpace(jnp.log(0.5))

    def evidence_chain(key):
        # T ~ Beta(n[i],1) <==> T ~ Kumaraswamy(n[i],1)
        log_T = jnp.log(random.uniform(key, n_i.shape, dtype=L_mid.dtype)) / n_i
        # log_T = jnp.where(n_i == 0., -jnp.inf, log_T)
        # log_T = jnp.log(random.beta(key, state.sample_collection.num_live_points, 1.))
        T = LogSpace(log_T)
        X = LogSpace(jnp.asarray([0.], log_L_samples.dtype)).concatenate(T).cumprod()
        dX = (X[:-1] - X[1:]).abs()
        dZ = dX * L_mid
        dZ = LogSpace(jnp.where(n_i == 0., -jnp.inf, dZ.log_abs_val))
        # dZ = LogSpace(jnp.where(jnp.isnan(dZ.log_abs_val), -jnp.inf, dZ.log_abs_val))
        Z = dZ.sum()
        # ESS = Z.square() / dZ.square().sum()
        return Z.log_abs_val

    log_Z_chains = vmap(evidence_chain)(random.split(key, S))
    Z_chains = LogSpace(log_Z_chains)
    return Z_chains.log_abs_val


def analytic_log_evidence(model: Model, S: int = 60):
    """
    Compute the evidence with brute-force over a regular grid.

    Args:
        prior_chain: PriorChain of model
        log_likelihood (:code:`callable(**samples)`): log-likelihood function
        S (int): resolution of grid

    Returns:
        log(Z)
    """

    u_vec = jnp.linspace(jnp.finfo(float_type).eps, 1. - jnp.finfo(float_type).eps, S)
    du = u_vec[1] - u_vec[0]
    args = jnp.stack([x.flatten() for x in jnp.meshgrid(*[u_vec] * model.U_ndims, indexing='ij')], axis=-1)
    Z_true = (LogSpace(jit(vmap(model.forward))(args)).nansum() * LogSpace(
        jnp.log(du)) ** model.U_ndims)
    return Z_true.log_abs_val


def analytic_posterior_samples(model: Model, S: int = 60):
    """
    Compute the evidence with brute-force over a regular grid.

    Args:
        prior_chain (jaxns.PriorChain): PriorChain of model
        log_likelihood (:code:`callable(**samples)`): log-likelihood function
        S (int): resolution of grid

    Returns:
        log(Z)
    """

    u_vec = jnp.linspace(jnp.finfo(float_type).eps, 1. - jnp.finfo(float_type).eps, S)
    du = u_vec[1] - u_vec[0]
    args = jnp.stack([x.flatten() for x in jnp.meshgrid(*[u_vec] * model.U_ndims, indexing='ij')], axis=-1)
    samples = jit(vmap(model.transform))(args)
    log_L = jit(vmap(model.forward))(args)
    dZ = LogSpace(log_L) * LogSpace(jnp.log(du)) ** model.U_ndims
    return samples, dZ.log_abs_val


def _isinstance_namedtuple(obj) -> bool:
    return (
            isinstance(obj, tuple) and
            hasattr(obj, '_asdict') and
            hasattr(obj, '_fields')
    )


def save_pytree(pytree: NamedTuple, save_file: str):
    """
    Saves results of nested sampler in a npz file.

    Args:
        results (NestedSamplerResults): Nested sampler result
        save_file (str): filename
    """
    pytree_np = tree_map(lambda v: np.asarray(v) if v is not None else None, pytree)

    def _pytree_asdict(pytree):
        _data_dict = pytree._asdict()
        data_dict = {}
        for k, v in _data_dict.items():
            if _isinstance_namedtuple(v):
                data_dict[k] = _pytree_asdict(v)
            elif isinstance(v, (dict, np.ndarray, None.__class__)):
                data_dict[k] = v
            else:
                raise ValueError("key, value pair {}, {} unknown".format(k, v))
        return data_dict

    data_dict = _pytree_asdict(pytree_np)
    np.savez(save_file, **data_dict)


def save_results(results: NestedSamplerResults, save_file: str):
    """
    Saves results of nested sampler in a npz file.

    Args:
        results (NestedSamplerResults): Nested sampler result
        save_file (str): filename
    """
    save_pytree(results, save_file)


def load_pytree(save_file: str):
    """
    Loads saved nested sampler results from a npz file.

    Args:
        save_file (str): filename

    Returns:
        NestedSamplerResults
    """
    _data_dict = np.load(save_file, allow_pickle=True)
    data_dict = {}
    for k, v in _data_dict.items():
        if v.size == 1:
            if v.item() is None:
                data_dict[k] = None
            else:
                data_dict[k] = tree_map(lambda v: jnp.asarray(v), v.item())
        else:
            data_dict[k] = jnp.asarray(v)

    return data_dict


def load_results(save_file: str) -> NestedSamplerResults:
    """
    Loads saved nested sampler results from a npz file.

    Args:
        save_file (str): filename

    Returns:
        NestedSamplerResults
    """
    data_dict = load_pytree(save_file)
    return NestedSamplerResults(**data_dict)



import tensorflow_probability.substrates.jax as tfp
from jax import numpy as jnp, random, tree_map, disable_jit

from jaxns import PriorModelGen, Prior, Model
from jaxns.initial_state import get_uniform_init_live_points
from jaxns.likelihood_samplers.multi_ellipsoid_utils import log_ellipsoid_volume, ellipsoid_clustering
from jaxns.types import float_type

tfpd = tfp.distributions


def test_ellipsoid_clustering():
    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=2))
        y = yield Prior(tfpd.Normal(loc=0, scale=2))

        return x, y

    def log_likelihood(x,y):
        return jnp.log(jnp.exp(-0.5*((x-0.5)/0.1)**2-0.5*((y-0.5)/0.1)**2) + jnp.exp(-0.5*((x-1.5)/0.1)**2-0.5*((y-1.5)/0.1)**2))
    model = Model(prior_model=prior_model,
                  log_likelihood=log_likelihood)

    n = 1000
    live_points = get_uniform_init_live_points(random.PRNGKey(43),
                                               num_live_points=n,
                                               model=model)
    keep = live_points.reservoir.log_L > log_likelihood(1.1, 1.1)
    reservoir = tree_map(lambda x: x[keep], live_points.reservoir)
    # import pylab as plt
    # plt.scatter(reservoir.point_U[:,0], reservoir.point_U[:,1])
    # plt.show()
    with disable_jit():
        params = ellipsoid_clustering(random.PRNGKey(42), points=reservoir.point_U,
                                  log_VS=jnp.asarray(0., float_type),
                                  max_num_ellipsoids=10)


def test_log_ellipsoid_volume():
    radii = jnp.ones(2)
    assert jnp.isclose(log_ellipsoid_volume(radii), jnp.log(jnp.pi))
    radii = jnp.ones(3)
    assert jnp.isclose(log_ellipsoid_volume(radii), jnp.log(4. * jnp.pi / 3.))



from functools import partial
from typing import Tuple, Optional, NamedTuple

from etils.array_types import PRNGKey, FloatArray, IntArray, BoolArray
from jax import core, numpy as jnp, tree_map, random, jit
from jax._src.lax.control_flow import while_loop

from jaxns.internals.maps import chunked_pmap
from jaxns.internals.stats import linear_to_log_stats
from jaxns.model import Model
from jaxns.slice_sampler import UniDimSliceSampler, SeedPoint, PreprocessType, AbstractSliceSampler
from jaxns.statistics import analyse_sample_collection
from jaxns.types import NestedSamplerState, Reservoir, int_type, float_type
from jaxns.utils import sort_samples

__all__ = ['AdaptiveRefinement']


class AdaptiveRefinement:
    def __init__(self, model: Model, uncert_improvement_patience: int, num_slices: int, num_parallel_samplers: int = 1,
                 slice_sampler: Optional[AbstractSliceSampler] = None):
        if uncert_improvement_patience <= 0:
            raise ValueError(f"uncert_improvement_patient should be > 0, got {uncert_improvement_patience}.")
        self.uncert_improvement_patience = uncert_improvement_patience
        self.num_parallel_samplers = num_parallel_samplers
        if slice_sampler is None:
            slice_sampler = UniDimSliceSampler(
                model=model,
                midpoint_shrink=True,
                multi_ellipse_bound=False
            )
        self.slice_sampler = slice_sampler
        self.num_slices = num_slices

    def split_state(self, state: NestedSamplerState) -> Tuple[NestedSamplerState, Reservoir]:
        """
        Splice a state into iid samples and non-iid samples.

        Note: must not be run in a JIT-context, as the nonzero operation non-static.

        Args:
            state: state to split

        Returns:
            a state of iid samples, and reservoir of non-iid samples
        """
        if isinstance(state.sample_collection.reservoir.iid, core.Tracer):
            raise RuntimeError("Tracer detected, but expected imperative context.")
        num_samples = state.sample_collection.reservoir.log_L.size
        (update_indicies,) = jnp.nonzero(
            jnp.bitwise_not(state.sample_collection.reservoir.iid) & (
                    jnp.arange(num_samples) < state.sample_collection.sample_idx
            )
        )

        (keep_indicies,) = jnp.nonzero(state.sample_collection.reservoir.iid)

        update_reservoir = tree_map(lambda x: x[update_indicies], state.sample_collection.reservoir)
        keep_reservoir = tree_map(lambda x: x[keep_indicies], state.sample_collection.reservoir)
        state = state._replace(
            sample_collection=state.sample_collection._replace(
                sample_idx=keep_indicies.size,
                reservoir=keep_reservoir
            )
        )
        return state, update_reservoir

    def combine_state(self, state: NestedSamplerState, update_reservoir: Reservoir) -> NestedSamplerState:
        """
        Concatenates a state and reservoir and sorts the state.

        Args:
            state: state
            update_reservoir: reservoir

        Returns:
            merged and sorted state
        """
        # concatenate reservoirs
        reservoir = tree_map(lambda x, y: jnp.concatenate([x, y], axis=0),
                             state.sample_collection.reservoir, update_reservoir)

        sample_collection = state.sample_collection._replace(
            sample_idx=state.sample_collection.sample_idx + update_reservoir.iid.size,
            reservoir=reservoir
        )
        # Sort
        sample_collection = sort_samples(sample_collection)
        state = state._replace(sample_collection=sample_collection)
        return state

    def _single_improvement(self, key: PRNGKey, seed_point: SeedPoint, log_L_constraint: FloatArray,
                            preprocess_data: PreprocessType) -> Reservoir:
        """
        Perform perfect slice sampling, thereby improving the sample.

        Args:
            key: PRNGKey
            seed_point: seed point to sample from
            log_L_constraint: the constraint to sample inside
            preprocess_data: preprocessing data

        Returns:
            a sample from the seed point sampled within the given constraint
        """
        sample = self.slice_sampler.get_sample(key=key,
                                               seed_point=seed_point,
                                               log_L_constraint=log_L_constraint,
                                               num_slices=self.num_slices,
                                               preprocess_data=preprocess_data)
        return Reservoir(*sample)

    @partial(jit, static_argnums=0)
    def improve(self, state: NestedSamplerState, iid_state: NestedSamplerState, non_iid_reservoir: Reservoir):
        """
        Performs perfect slice sampling on the set of samples that are not considered i.i.d. sampled within their
        respective likelihood constraint.

        Args:
            state: state to improve
            iid_state: state with only the i.i.d. samples
            non_iid_reservoir: samples that are not deemed i.i.d.

        Returns:
            state where all samples are considered i.i.d. sampled within their respective likelihood constraints
        """

        # update each sample one slice at a time until the stopping condition

        class CarryType(NamedTuple):
            key: PRNGKey
            reservoir: Reservoir
            last_log_Z_mean: FloatArray
            last_diff_log_Z_mean: FloatArray
            patience: IntArray
            preprocess_data: PreprocessType

        def cond(body_state: CarryType) -> BoolArray:
            (_, _, _, _, patience, _) = body_state
            done = jnp.greater_equal(patience, self.uncert_improvement_patience)
            return jnp.bitwise_not(done)

        def body(body_state: CarryType) -> CarryType:
            batch_size = body_state.reservoir.iid.size

            key, combine_key, improve_key = random.split(body_state.key, 3)

            parallel_improvement = chunked_pmap(
                lambda key, seed_point, log_L_constraint: self._single_improvement(
                    key=key,
                    seed_point=seed_point,
                    log_L_constraint=log_L_constraint,
                    preprocess_data=body_state.preprocess_data
                ),
                chunksize=self.num_parallel_samplers,
                batch_size=batch_size)
            seed_points = SeedPoint(
                U0=body_state.reservoir.point_U,
                log_L0=body_state.reservoir.log_L
            )
            update_reservoir = parallel_improvement(random.split(improve_key, batch_size),
                                                    seed_points,
                                                    body_state.reservoir.log_L_constraint)

            update_reservoir = update_reservoir._replace(
                num_slices=update_reservoir.num_slices + body_state.reservoir.num_slices,
                num_likelihood_evaluations=(update_reservoir.num_likelihood_evaluations
                                            + body_state.reservoir.num_likelihood_evaluations)
            )

            updated_state = self.combine_state(state=iid_state, update_reservoir=update_reservoir)
            evidence_calculation, sample_stats = analyse_sample_collection(
                sample_collection=updated_state.sample_collection,
                sorted_collection=True
            )
            log_Z_mean, log_Z_var = linear_to_log_stats(
                log_f_mean=evidence_calculation.log_Z_mean,
                log_f2_mean=evidence_calculation.log_Z2_mean
            )
            nan_last_log_Z_mean = jnp.isnan(body_state.last_log_Z_mean)
            nan_log_Z_var = jnp.isnan(log_Z_var)
            nan_log_Z_mean = jnp.isnan(log_Z_mean)

            diff_log_Z_mean = jnp.abs(body_state.last_log_Z_mean - log_Z_mean)
            small_change = jnp.square(2. * diff_log_Z_mean) < log_Z_var
            decreasing_change = diff_log_Z_mean < body_state.last_diff_log_Z_mean
            stable = decreasing_change | small_change
            reset_patience = jnp.bitwise_not(stable) | (nan_log_Z_mean | nan_last_log_Z_mean | nan_log_Z_var)
            patience = jnp.where(reset_patience, jnp.zeros_like(body_state.patience),
                                 body_state.patience + jnp.ones_like(body_state.patience))
            preprocess_data = self.slice_sampler.preprocess(updated_state)
            return CarryType(key, update_reservoir, log_Z_mean, diff_log_Z_mean, patience, preprocess_data)

        preprocess_data = self.slice_sampler.preprocess(state)

        key, improve_key = random.split(state.key, 2)
        init_evidence_calculation, _ = analyse_sample_collection(
            sample_collection=state.sample_collection,
            sorted_collection=True
        )
        init_log_Z_mean, _ = linear_to_log_stats(
            log_f_mean=init_evidence_calculation.log_Z_mean,
            log_f2_mean=init_evidence_calculation.log_Z2_mean
        )

        init_body_state: CarryType = CarryType(
            improve_key, non_iid_reservoir, init_log_Z_mean, jnp.asarray(0., float_type), jnp.asarray(0, int_type),
            preprocess_data)
        output_state = while_loop(cond,
                                  body,
                                  init_body_state)

        # mark as iid now <==> the evidence stopped changing
        update_reservoir = output_state.reservoir._replace(iid=jnp.ones_like(output_state.reservoir.iid))

        updated_state = self.combine_state(state=iid_state, update_reservoir=update_reservoir)

        updated_state = updated_state._replace(key=key)
        return updated_state

    def __call__(self, state: NestedSamplerState) -> NestedSamplerState:
        if isinstance(state.sample_collection.reservoir.iid, core.Tracer):
            raise RuntimeError("Tracer detected, but expected imperative context.")
        iid_state, non_iid_reservoir = self.split_state(state)
        return self.improve(state=state, iid_state=iid_state, non_iid_reservoir=non_iid_reservoir)



from typing import Tuple, Union

from etils.array_types import FloatArray, IntArray
from jax import numpy as jnp, tree_map, random
from jax._src.lax.control_flow import while_loop
from jax._src.lax.slicing import dynamic_update_slice

from jaxns.internals.log_semiring import LogSpace, normalise_log_space
from jaxns.internals.maps import replace_index
from jaxns.internals.stats import linear_to_log_stats
from jaxns.types import EvidenceCalculation, SampleCollection, SampleStatistics, float_type, int_type
from jaxns.types import NestedSamplerState

__all__ = ['compute_evidence',
           'analyse_sample_collection',
           'compute_num_live_points_from_unit_threads']


def _init_evidence_calculation() -> EvidenceCalculation:
    """
    Returns an initial evidence calculation object.

    Returns:
        EvidenceCalculation
    """
    evidence_calculation = EvidenceCalculation(
        log_X_mean=jnp.asarray(0., float_type),
        log_X2_mean=jnp.asarray(0., float_type),
        log_Z_mean=jnp.asarray(-jnp.inf, float_type),
        log_ZX_mean=jnp.asarray(-jnp.inf, float_type),
        log_Z2_mean=jnp.asarray(-jnp.inf, float_type),
        log_dZ2_mean=jnp.asarray(-jnp.inf, float_type)
    )
    return evidence_calculation


def _update_evidence_calculation(num_live_points: FloatArray, log_L: FloatArray, next_log_L_contour: FloatArray,
                                 evidence_calculation: EvidenceCalculation) -> EvidenceCalculation:
    """
    Update an evidence calculation with the next sample from the shrinkage distribution.

    Args:
        num_live_points: number of live points at the current shrinkage step
        log_L: the log likelihood of the current contour
        next_log_L_contour: the log likelihood of the next contour after shrinkage
        evidence_calculation: the evidence calculation to update

    Returns:
        an updated evidence calculation
    """
    # num_live_points = num_live_points.astype(float_type)
    next_L = LogSpace(next_log_L_contour)
    L_contour = LogSpace(log_L)
    midL = LogSpace(jnp.log(0.5)) * (next_L + L_contour)
    X_mean = LogSpace(evidence_calculation.log_X_mean)
    X2_mean = LogSpace(evidence_calculation.log_X2_mean)
    Z_mean = LogSpace(evidence_calculation.log_Z_mean)
    ZX_mean = LogSpace(evidence_calculation.log_ZX_mean)
    Z2_mean = LogSpace(evidence_calculation.log_Z2_mean)
    dZ2_mean = LogSpace(evidence_calculation.log_dZ2_mean)

    # T_mean = LogSpace(jnp.log(num_live_points) - jnp.log(num_live_points + 1.))
    # T_mean = LogSpace(jnp.log(1.) - jnp.log(1. + 1./num_live_points))
    T_mean = LogSpace(- jnp.logaddexp(0., -jnp.log(num_live_points)))
    # T_mean = LogSpace(- jnp.logaddexp(0., -jnp.log(num_live_points)))
    t_mean = LogSpace(- jnp.log(num_live_points + 1.))
    # T2_mean = LogSpace(jnp.log(num_live_points) - jnp.log( num_live_points + 2.))
    # T2_mean = LogSpace(jnp.log(1.) - jnp.log(1. + 2./num_live_points))
    T2_mean = LogSpace(- jnp.logaddexp((0.), jnp.log(2.) - jnp.log(num_live_points)))
    # T2_mean = LogSpace(- jnp.logaddexp(jnp.log(2.), -jnp.log(num_live_points)))
    t2_mean = LogSpace(jnp.log(2.) - jnp.log(num_live_points + 1.) - jnp.log(num_live_points + 2.))
    # tT_mean = LogSpace(jnp.log(num_live_points) - jnp.log(num_live_points + 1.) - jnp.log(num_live_points + 2.))
    # tT_mean = LogSpace(jnp.log(1.) - jnp.log(1. + 1./num_live_points) - jnp.log(num_live_points + 2.))
    tT_mean = LogSpace(- jnp.logaddexp(0., -jnp.log(num_live_points)) - jnp.log(num_live_points + 2.))
    # tT_mean = LogSpace(- jnp.logaddexp(0., -jnp.log(num_live_points)) - jnp.log(num_live_points + 2.))

    next_X_mean = X_mean * T_mean
    next_X2_mean = X2_mean * T2_mean
    next_Z_mean = Z_mean + X_mean * t_mean * midL
    next_ZX_mean = ZX_mean * T_mean + X2_mean * tT_mean * midL
    next_Z2_mean = Z2_mean + LogSpace(jnp.log(2.)) * ZX_mean * t_mean * midL + (X2_mean * t2_mean * midL ** 2)
    next_dZ2_mean = dZ2_mean + (X2_mean * t2_mean * midL ** 2)

    next_evidence_calculation = evidence_calculation._replace(
        log_X_mean=next_X_mean.log_abs_val.astype(float_type),
        log_X2_mean=next_X2_mean.log_abs_val.astype(float_type),
        log_Z_mean=next_Z_mean.log_abs_val.astype(float_type),
        log_Z2_mean=next_Z2_mean.log_abs_val.astype(float_type),
        log_ZX_mean=next_ZX_mean.log_abs_val.astype(float_type),
        log_dZ2_mean=next_dZ2_mean.log_abs_val.astype(float_type)
    )
    # log_Z_mean, log_Z_var = linear_to_log_stats(next_evidence_calculation.log_Z_mean, log_f2_mean=next_evidence_calculation.log_Z2_mean)

    # next_evidence_calculation = tree_map(lambda old, new: jnp.where(jnp.isnan(new), old, new),
    #                                      evidence_calculation, next_evidence_calculation)

    return next_evidence_calculation


def compute_evidence(sample_collection: SampleCollection, num_live_points: FloatArray) \
        -> Tuple[EvidenceCalculation, SampleStatistics]:
    """
    Compute the evidence by traversing the sample collection.

    Args:
        sample_collection: the sorted sample collection to use to calculate the evidence
        num_live_points: the number of live points at each sample (see compute_num_live_points_from_unit_threads)

    Returns:
        evidence calculation and the sample statistics
    """
    num_samples = sample_collection.sample_idx

    CarryType = Tuple[EvidenceCalculation, IntArray, FloatArray, FloatArray, FloatArray]

    def thread_cond(body_state: CarryType):
        (evidence_calculation, idx, log_L_contour, log_dZ_mean, log_X_mean) = body_state
        return idx < num_samples

    def thread_body(body_state: CarryType) -> CarryType:
        (evidence_calculation, idx, log_L_contour, log_dZ_mean, log_X_mean) = body_state
        next_num_live_points = num_live_points[idx]
        next_log_L = sample_collection.reservoir.log_L[idx]
        next_log_L_contour = next_log_L
        # Get log_dZ_mean, and log_X_mean
        next_evidence_calculation = _update_evidence_calculation(
            num_live_points=next_num_live_points,
            log_L=log_L_contour,
            next_log_L_contour=next_log_L,
            evidence_calculation=evidence_calculation
        )

        next_dZ_mean = (LogSpace(next_evidence_calculation.log_Z_mean)
                        - LogSpace(evidence_calculation.log_Z_mean)).abs()

        next_log_dZ_mean = log_dZ_mean.at[idx].set(next_dZ_mean.log_abs_val)
        next_log_X_mean = log_X_mean.at[idx].set(next_evidence_calculation.log_X_mean)
        next_idx = idx + jnp.ones_like(idx)
        return (next_evidence_calculation, next_idx, next_log_L_contour, next_log_dZ_mean, next_log_X_mean)

    initial_evidence_calculation = _init_evidence_calculation()
    init_log_L_contour = sample_collection.reservoir.log_L_constraint[0]
    init_log_dZ_mean = jnp.full(num_live_points.shape, -jnp.inf, float_type)
    init_log_X_mean = jnp.full(num_live_points.shape, -jnp.inf, float_type)

    (final_evidence_calculation, final_idx, final_log_L_contour, final_log_dZ_mean, final_log_X_mean) = \
        while_loop(thread_cond,
                   thread_body,
                   (initial_evidence_calculation, jnp.asarray(0, int_type), init_log_L_contour,
                    init_log_dZ_mean, init_log_X_mean))

    sample_stats = SampleStatistics(
        num_live_points=num_live_points,
        log_X_mean=final_log_X_mean,
        log_dZ_mean=final_log_dZ_mean,
    )
    return final_evidence_calculation, sample_stats


def analyse_sample_collection(sample_collection: SampleCollection, sorted_collection: bool = True) \
        -> Tuple[EvidenceCalculation, SampleStatistics]:
    """
    Computes the evidence and statistics.

    Args:
        sample_collection: the sample collection
        sorted_collection: whether the sample collection is sorted

    Returns:
        evidence calculation and sample stats
    """
    # Sample collection has unsorted samples, and incorrect num_live_points.
    if sorted_collection:
        num_live_points = compute_num_live_points_from_unit_threads(
            log_L_constraints=sample_collection.reservoir.log_L_constraint,
            log_L_samples=sample_collection.reservoir.log_L,
            num_samples=sample_collection.sample_idx,
            sorted_collection=True
        )
    else:
        num_live_points, sort_idx = compute_num_live_points_from_unit_threads(
            log_L_constraints=sample_collection.reservoir.log_L_constraint,
            log_L_samples=sample_collection.reservoir.log_L,
            num_samples=sample_collection.sample_idx,
            sorted_collection=False
        )

        sample_collection = sample_collection._replace(
            reservoir=tree_map(lambda x: x[sort_idx], sample_collection.reservoir)
        )

    evidence_calculation, sample_stats = compute_evidence(
        sample_collection=sample_collection,
        num_live_points=num_live_points
    )

    return evidence_calculation, sample_stats


def compute_num_live_points_from_unit_threads(log_L_constraints: FloatArray, log_L_samples: FloatArray,
                                              num_samples: IntArray = None, sorted_collection: bool = True) \
        -> Union[FloatArray, Tuple[FloatArray, IntArray]]:
    """
    Compute the number of live points of shrinkage distribution, from an arbitrary list of samples with
    corresponding sampling constraints.

    Args:
        log_L_constraints: [N] likelihood constraint that sample was uniformly sampled within
        log_L_samples: [N] likelihood of the sample
        sorted_collection: bool, whether the sample collection was already sorted.

    Returns:
        if sorted_collection is true:
            num_live_points for shrinkage distribution
        otherwise:
            num_live_points for shrinkage distribution, and sort indicies
    """
    # mask the samples that are not yet taken

    if num_samples is None:
        num_samples = log_L_samples.size
    empty_mask = jnp.greater_equal(jnp.arange(log_L_samples.size), num_samples)

    # masking samples is already done, since they are inf by default.
    # log_L_samples = jnp.where(empty_mask, jnp.inf, log_L_samples)
    # log_L_constraints = jnp.where(empty_mask, jnp.inf, log_L_constraints)
    # sort log_L_samples, breaking degeneracies on log_L_constraints

    if not sorted_collection:
        idx_sort = jnp.lexsort((log_L_constraints, log_L_samples))
        log_L_samples = log_L_samples[idx_sort]
        log_L_constraints = log_L_constraints[idx_sort]
        empty_mask = empty_mask[idx_sort]

    # n = jnp.sum(available & (log_L_samples >= contour) & (log_L_constraints <= contour))
    # n = jnp.sum(available & (log_L_constraints <= contour))
    # n = available.size - jnp.sum(jnp.bitwise_not(available) | jnp.bitwise_not(log_L_constraints <= contour))
    # n = available.size - jnp.sum(jnp.bitwise_not(available)) - jnp.sum(jnp.bitwise_not(log_L_constraints <= contour)) + jnp.sum(jnp.bitwise_not(available) & jnp.bitwise_not(log_L_constraints <= contour))
    # n = jnp.sum(available) - jnp.sum(log_L_constraints > contour) + jnp.sum(jnp.bitwise_not(available) & (log_L_constraints > contour))
    # Since jnp.sum(jnp.bitwise_not(available) & (log_L_constraints > contour)) can be shown to be zero
    # n = jnp.sum(available) - jnp.sum(log_L_constraints > contour)

    _log_L_samples = jnp.concatenate([log_L_constraints[0:1], log_L_samples])
    n_base = jnp.arange(1, log_L_samples.size + 1)[::-1]

    # jnp.sum(log_L_samples[idx-1] < log_L_constraints)

    u = jnp.searchsorted(_log_L_samples, log_L_constraints, side='left')
    b = u.size - jnp.cumsum(jnp.bincount(u, length=u.size))
    n = n_base - b
    n = jnp.where(empty_mask, 0, n)
    n = jnp.asarray(n, log_L_samples.dtype)
    if not sorted_collection:
        return (n, idx_sort)
    return n


def sample_goal_distribution(key, log_goal_weights, S: int, *, replace: bool = True):
    """
    Sample indices that match unnormalised log_probabilities.

    Args:
        key: PRNG key
        log_goal_weights: unnormalised log probabilities
        S: number of samples
        replace: bool, whether to sample with replacement

    Returns:
        indices that draw from target density
    """
    if replace:
        p_cuml = LogSpace(log_goal_weights).cumsum()
        # 1 - U in (0,1] instead of [0,1)
        r = p_cuml[-1] * LogSpace(jnp.log(1 - random.uniform(key, (S,))))
        idx = jnp.searchsorted(p_cuml.log_abs_val, r.log_abs_val)
    else:
        assert S <= log_goal_weights.size
        g = -random.gumbel(key, shape=log_goal_weights.shape) - log_goal_weights
        idx = jnp.argsort(g)[:S]
    return idx


def compute_remaining_evidence(sample_idx, log_dZ_mean):
    # Z_remaining = dZ_mean.cumsum(reverse=True)

    def logsumexp_cumsum_body(_state):
        (log_abs_val, idx) = _state
        next_idx = idx - jnp.ones_like(idx)
        next_val = LogSpace(log_abs_val[idx]) + LogSpace(log_abs_val[next_idx])
        next_log_abs_val = dynamic_update_slice(log_abs_val, next_val.log_abs_val[None], [next_idx])
        return (next_log_abs_val, next_idx)

    # Calculate remaining evidence, doing only the minimal amount of work necessary.

    (log_Z_remaining, _) = while_loop(lambda _state: _state[1] > 0,
                                      logsumexp_cumsum_body,
                                      (log_dZ_mean,
                                       sample_idx - jnp.ones_like(sample_idx)))
    return log_Z_remaining


def evidence_goal(state: NestedSamplerState):
    """
    Estimates the impact of adding a sample at a certain likelihood contour by computing the impact of removing a point.

    Args:
        state:

    Returns:

    """
    # evidence uncertainty minimising goal.
    # remove points and see what increases uncertainty the most.

    _, log_Z_var0 = linear_to_log_stats(log_f_mean=state.evidence_calculation.log_Z_mean,
                                        log_f2_mean=state.evidence_calculation.log_Z2_mean)
    num_shrinkages = -state.evidence_calculation.log_X_mean
    delta_idx = state.sample_idx / (2. * num_shrinkages)

    def body(body_state):
        (remove_idx, inf_max_dvar, inf_max_dvar_idx, sup_max_dvar, sup_max_dvar_idx) = body_state
        # removing a sample is equivalent to setting that n=inf at that point
        perturbed_num_live_points = replace_index(state.sample_collection.num_live_points, jnp.inf,
                                                  remove_idx.astype(int_type))
        perturbed_sample_collection = state.sample_collection._replace(num_live_points=perturbed_num_live_points)
        (evidence_calculation, _, _, _, _) = \
            compute_evidence(num_samples=state.sample_idx, sample_collection=perturbed_sample_collection)
        _, log_Z_var = linear_to_log_stats(log_f_mean=evidence_calculation.log_Z_mean,
                                           log_f2_mean=evidence_calculation.log_Z2_mean)
        dvar = log_Z_var - log_Z_var0

        return remove_idx + delta_idx, dvar


def _get_dynamic_goal(state: NestedSamplerState, G: jnp.ndarray):
    """
    Get contiguous contours that we'd like to reinforce.

    We have two objectives, which can be mixed by setting `G`.
    G=0: choose contours that decrease evidence uncertainty the most.
    G=1: choose contours that increase ESS the most.

    Note: This slightly departs from the Dynamic Nested Sampling paper.
    """

    n_i = state.sample_collection.num_live_points
    dZ_mean = LogSpace(state.sample_collection.log_dZ_mean)
    # Calculate remaining evidence, doing only the amount of work necessary.
    Z_remaining = LogSpace(compute_remaining_evidence(state.sample_idx, state.sample_collection.log_dZ_mean))
    # TODO: numerically compute goal using custome norm
    I_evidence = ((LogSpace(jnp.log(n_i + 1.)) * Z_remaining + LogSpace(jnp.log(n_i)) * dZ_mean) / (
            LogSpace(jnp.log(n_i + 1.)).sqrt() * LogSpace(jnp.log(n_i + 2.)) ** (1.5)))
    I_evidence = normalise_log_space(I_evidence)

    I_posterior = dZ_mean
    I_posterior = normalise_log_space(I_posterior)

    I_goal = LogSpace(jnp.log(1. - G)) * I_evidence + LogSpace(jnp.log(G)) * I_posterior
    # I_goal = normalise_log_space(I_goal) # unnecessary for sampling

    mask = jnp.arange(I_goal.size) >= state.sample_idx
    I_goal = LogSpace(jnp.where(mask, -jnp.inf, I_goal.log_abs_val))

    return I_goal.log_abs_val


def get_dynamic_goal(key, state: NestedSamplerState, num_samples: int, G: jnp.ndarray) -> Tuple[
    jnp.ndarray, jnp.ndarray]:
    """
    Determines what seed points to sample above.
    """
    contours = jnp.concatenate([state.sample_collection.log_L_constraint[0:1],
                                state.sample_collection.log_L_samples])
    if G is None:
        raise ValueError(f"G should be a float in [0,1].")
    log_goal_weights = _get_dynamic_goal(state, G)
    # Probabilistically sample the contours according to goal distribution
    indices_constraint_reinforce = sample_goal_distribution(key, log_goal_weights, num_samples, replace=True)
    start_idx = indices_constraint_reinforce.min()
    end_idx = indices_constraint_reinforce.max()
    log_L_constraint_start = contours[start_idx]
    log_L_constraint_end = contours[end_idx]

    return log_L_constraint_start, log_L_constraint_end



from typing import Tuple

from etils.array_types import BoolArray
from jax import random, numpy as jnp, tree_map
from jax.lax import while_loop

from jaxns.internals.maps import replace_index
from jaxns.model import Model
from jaxns.types import Sample, NestedSamplerState, LivePoints, Reservoir
from jaxns.uniform_sampler import UniformSampler

__all__ = ['StaticUniform']


class StaticUniform:
    def __init__(self, model: Model, num_live_points: int, efficiency_threshold: float):
        self.uniform_sampler = UniformSampler(model=model)
        self.model = model
        self.num_live_points = num_live_points
        if efficiency_threshold <= 0.:
            raise ValueError(f"Efficiency threshold should be > 0, got {efficiency_threshold}.")
        self.efficiency_threshold = efficiency_threshold

    def __call__(self, state: NestedSamplerState, live_points: LivePoints) -> Tuple[NestedSamplerState, LivePoints]:
        if live_points.reservoir.log_L.size != self.num_live_points:
            raise ValueError(
                f"live points reservoir is the wrong size. "
                f"Got {live_points.reservoir.log_L.size} by expected {self.num_live_points}.")

        CarryType = Tuple[BoolArray, NestedSamplerState, LivePoints]

        def body(carry: CarryType) -> CarryType:
            (_, state, live_points) = carry
            idx_min = jnp.argmin(live_points.reservoir.log_L)
            dead_point: Sample = tree_map(lambda x: x[idx_min], live_points.reservoir)

            sample_collection_reservoir = tree_map(
                lambda old, update: replace_index(old, update, state.sample_collection.sample_idx),
                state.sample_collection.reservoir, dead_point)

            sample_collection = state.sample_collection._replace(reservoir=sample_collection_reservoir,
                                                                 sample_idx=state.sample_collection.sample_idx + 1)

            log_L_dead = dead_point.log_L

            # contour becomes log_L_dead if log_L_dead is not supremum of live-points, else we choose the original
            # constraint of dead point. Note: we are at liberty to choose any log_L level as a contour so long as we
            # can sample within it uniformly.
            on_supremum = jnp.equal(log_L_dead, jnp.max(live_points.reservoir.log_L))
            log_L_contour = jnp.where(on_supremum, dead_point.log_L_constraint, log_L_dead)

            # replace dead point with a new sample about contour
            key, sample_key = random.split(state.key, 2)
            sample = self.uniform_sampler.single_sample(key=sample_key, log_L_constraint=log_L_contour)
            live_points_reservoir = tree_map(lambda old, update: replace_index(old, update, idx_min),
                                             live_points.reservoir, Reservoir(*sample))
            live_points = live_points._replace(reservoir=live_points_reservoir)
            state = state._replace(key=key,
                                   sample_collection=sample_collection)
            # done = sample.num_likelihood_evaluations >= 1. / self.efficiency_threshold
            # done = jnp.mean(live_points.reservoir.num_likelihood_evaluations) >= 1. / self.efficiency_threshold
            on_plateau = jnp.min(live_points_reservoir.log_L) == jnp.max(live_points_reservoir.log_L)
            sufficient_compression = sample_collection.sample_idx / self.num_live_points > -jnp.log(
                self.efficiency_threshold)  # -log(X) = num_samples/num_live_points = -log(eff)
            done = on_plateau | sufficient_compression
            return (done, state, live_points)

        (_, state, live_points) = while_loop(lambda carry: jnp.bitwise_not(carry[0]),
                                             body, (jnp.asarray(False, jnp.bool_), state, live_points))
        return (state, live_points)



from time import monotonic_ns

import tensorflow_probability.substrates.jax as tfp
from jax import random, numpy as jnp
from jax._src.scipy.linalg import solve_triangular

from jaxns.model import Model
from jaxns.nested_sampler import ApproximateNestedSampler, ExactNestedSampler
from jaxns.prior import PriorModelGen, Prior
from jaxns.slice_sampler import MultiDimSliceSampler, UniDimSliceSampler
from jaxns.types import TerminationCondition
from jaxns.utils import evidence_posterior_samples, analytic_log_evidence

tfpd = tfp.distributions


class Timer:
    def __enter__(self):
        self.t0 = monotonic_ns()

    def __exit__(self, exc_type, exc_val, exc_tb):
        print(f"Time to execute: {(monotonic_ns() - self.t0) / 1e9} seconds.")


def test_approximate_nested_sampler():
    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=jnp.asarray([1., 1.])))
        y = yield Prior(tfpd.Normal(loc=2, scale=x))
        z = x + y
        return z

    def log_likelihood(z):
        return -jnp.sum(z ** 2)

    model = Model(prior_model=prior_model,
                  log_likelihood=log_likelihood)
    approx_ns = ApproximateNestedSampler(model=model, num_live_points=50, num_parallel_samplers=1,
                                         max_samples=1000)
    # print(termination_reason)
    # print(state)
    with Timer():
        termination_reason, state = approx_ns(random.PRNGKey(42),
                                              term_cond=TerminationCondition(live_evidence_frac=1e-4))
        results = approx_ns.to_results(state, termination_reason)
        results.log_Z_mean.block_until_ready()

    with Timer():
        termination_reason, state = approx_ns(random.PRNGKey(42),
                                              term_cond=TerminationCondition(live_evidence_frac=1e-4))
        results = approx_ns.to_results(state, termination_reason)
        results.log_Z_mean.block_until_ready()


def test_exact_nested_sampler():
    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=2))
        y = yield Prior(tfpd.Normal(loc=2, scale=x))
        z = x + y
        return z

    def log_likelihood(z):
        return -z ** 2

    model = Model(prior_model=prior_model, log_likelihood=log_likelihood)
    model.sanity_check(random.PRNGKey(0), S=100)

    exact_ns = ExactNestedSampler(model=model, num_live_points=50, num_parallel_samplers=1, max_samples=1000,
                                  uncert_improvement_patience=1)
    # print(termination_reason)
    # print(state)
    with Timer():
        termination_reason, state = exact_ns(random.PRNGKey(42),
                                             term_cond=TerminationCondition(live_evidence_frac=1e-4))
        results = exact_ns.to_results(state, termination_reason)
        exact_ns.summary(results)
        exact_ns.plot_diagnostics(results)


    with Timer():
        termination_reason, state = exact_ns(random.PRNGKey(42),
                                             term_cond=TerminationCondition(live_evidence_frac=1e-4))
        results = exact_ns.to_results(state, termination_reason)


def test_exact_nested_sampler_multidim_slice_sampler():
    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=2))
        y = yield Prior(tfpd.Normal(loc=2, scale=x))
        z = x + y
        return z

    def log_likelihood(z):
        return -z ** 2

    model = Model(prior_model=prior_model,
                  log_likelihood=log_likelihood)
    slice_sampler = MultiDimSliceSampler(model=model, num_restrict_dims=model.U_ndims)
    exact_ns = ExactNestedSampler(model=model, num_live_points=50, num_parallel_samplers=1,
                                  max_samples=1000, refinement_slice_sampler=slice_sampler)
    # print(termination_reason)
    # print(state)
    with Timer():
        termination_reason, state = exact_ns(random.PRNGKey(42),
                                             term_cond=TerminationCondition(live_evidence_frac=1e-4))
        results = exact_ns.to_results(state, termination_reason)
        results.log_Z_mean.block_until_ready()

    with Timer():
        termination_reason, state = exact_ns(random.PRNGKey(42),
                                             term_cond=TerminationCondition(live_evidence_frac=1e-4))
        results = exact_ns.to_results(state, termination_reason)
        results.log_Z_mean.block_until_ready()


def test_nested_sampling_basic():
    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=1), name='x')
        return x

    def log_likelihood(x):
        return - jnp.sum(x ** 2)

    model = Model(prior_model=prior_model,
                  log_likelihood=log_likelihood)
    exact_ns = ExactNestedSampler(model=model, num_live_points=50, num_parallel_samplers=1,
                                  max_samples=1000)
    with Timer():
        termination_reason, state = exact_ns(random.PRNGKey(42),
                                             term_cond=TerminationCondition(live_evidence_frac=1e-4))
        results = exact_ns.to_results(state, termination_reason)

        # exact_ns.plot_diagnostics(results)
        exact_ns.summary(results)
        # exact_ns.plot_cornerplot(results)

    log_Z_true = analytic_log_evidence(model=model, S=200)

    log_Z_samples = evidence_posterior_samples(random.PRNGKey(42),
                                               results.num_live_points_per_sample,
                                               results.log_L_samples, S=1000)

    assert jnp.isclose(results.log_Z_mean, jnp.mean(log_Z_samples), atol=1e-3)
    assert jnp.isclose(results.log_Z_uncert, jnp.std(log_Z_samples), atol=1e-2)

    assert jnp.bitwise_not(jnp.isnan(results.log_Z_mean))
    assert jnp.isclose(results.log_Z_mean, log_Z_true, atol=1.75 * results.log_Z_uncert)


#
# def test_nested_sampling_basic_parallel():
#     def prior_model() -> PriorModelGen:
#         x = yield Prior(tfpd.Uniform(low=0, high=1), name='x')
#         return x
#
#     def log_likelihood(x):
#         return - jnp.sum(x ** 2)
#
#     model = Model(prior_model=prior_model,
#                   log_likelihood=log_likelihood)
#     exact_ns = ExactNestedSampler(model=model, num_live_points=50, num_parallel_samplers=2,
#                                   max_samples=1000)
#
#     termination_reason, state = exact_ns(random.PRNGKey(42),
#                                          term_cond=TerminationCondition(live_evidence_frac=1e-4))
#     results = exact_ns.to_results(state, termination_reason)
#
#     exact_ns.plot_diagnostics(results)
#     exact_ns.summary(results)
#     exact_ns.plot_cornerplot(results)
#
#     log_Z_true = analytic_log_evidence(model=model, S=200)
#
#     log_Z_samples = evidence_posterior_samples(random.PRNGKey(42),
#                                                results.num_live_points_per_sample,
#                                                results.log_L_samples, S=1000)
#
#     assert jnp.isclose(results.log_Z_mean, jnp.mean(log_Z_samples), atol=1e-3)
#     assert jnp.isclose(results.log_Z_uncert, jnp.std(log_Z_samples), atol=1e-3)
#
#     assert jnp.bitwise_not(jnp.isnan(results.log_Z_mean))
#     assert jnp.isclose(results.log_Z_mean, log_Z_true, atol=1.75 * results.log_Z_uncert)


def test_nested_sampling_mvn_static():
    def log_normal(x, mean, cov):
        L = jnp.linalg.cholesky(cov)
        dx = x - mean
        dx = solve_triangular(L, dx, lower=True)
        return -0.5 * x.size * jnp.log(2. * jnp.pi) - jnp.sum(jnp.log(jnp.diag(L))) \
               - 0.5 * dx @ dx

    ndims = 4
    prior_mu = 2 * jnp.ones(ndims)
    prior_cov = jnp.diag(jnp.ones(ndims)) ** 2

    data_mu = jnp.zeros(ndims)
    data_cov = jnp.diag(jnp.ones(ndims)) ** 2
    data_cov = jnp.where(data_cov == 0., 0.95, data_cov)

    true_logZ = log_normal(data_mu, prior_mu, prior_cov + data_cov)
    # not super happy with this being 1.58 and being off by like 0.1. Probably related to the ESS.
    post_mu = prior_cov @ jnp.linalg.inv(prior_cov + data_cov) @ data_mu + data_cov @ jnp.linalg.inv(
        prior_cov + data_cov) @ prior_mu

    print(f"True post mu:{post_mu}")
    print(f"True log Z: {true_logZ}")

    def prior_model() -> PriorModelGen:
        x = yield Prior(
            tfpd.MultivariateNormalTriL(loc=prior_mu, scale_tril=jnp.linalg.cholesky(prior_cov)),
            name='x')
        return x

    def log_likelihood(x):
        return log_normal(x, data_mu, data_cov)

    model = Model(prior_model=prior_model, log_likelihood=log_likelihood)

    model.sanity_check(random.PRNGKey(52), S=100)
    exact_ns = ExactNestedSampler(model=model, num_live_points=200, num_parallel_samplers=1,
                                  max_samples=3000)
    with Timer():
        termination_reason, state = exact_ns(random.PRNGKey(42),
                                             term_cond=TerminationCondition(live_evidence_frac=1e-4))
        results = exact_ns.to_results(state, termination_reason)
        exact_ns.summary(results)
    # exact_ns.plot_diagnostics(results)
    assert jnp.isclose(results.log_Z_mean, true_logZ, atol=1.75 * results.log_Z_uncert)


def test_multi_needs_fewer_slices():

    def log_normal(x, mean, cov):
        L = jnp.linalg.cholesky(cov)
        dx = x - mean
        dx = solve_triangular(L, dx, lower=True)
        return -0.5 * x.size * jnp.log(2. * jnp.pi) - jnp.sum(jnp.log(jnp.diag(L))) - 0.5 * dx @ dx

    ndims = 4
    prior_mu = 2 * jnp.ones(ndims)
    prior_cov = jnp.diag(jnp.ones(ndims)) ** 2

    data_mu = jnp.zeros(ndims)
    data_cov = jnp.diag(jnp.ones(ndims)) ** 2
    data_cov = jnp.where(data_cov == 0., 0.95, data_cov)

    true_logZ = log_normal(data_mu, prior_mu, prior_cov + data_cov)
    # not super happy with this being 1.58 and being off by like 0.1. Probably related to the ESS.
    post_mu = prior_cov @ jnp.linalg.inv(prior_cov + data_cov) @ data_mu + data_cov @ jnp.linalg.inv(
        prior_cov + data_cov) @ prior_mu

    print(f"True post mu:{post_mu}")
    print(f"True log Z: {true_logZ}")

    def prior_model() -> PriorModelGen:
        x = yield Prior(
            tfpd.MultivariateNormalTriL(loc=prior_mu, scale_tril=jnp.linalg.cholesky(prior_cov)),
            name='x')
        return x

    def log_likelihood(x):
        return log_normal(x, data_mu, data_cov)

    model = Model(prior_model=prior_model, log_likelihood=log_likelihood)

    model.sanity_check(random.PRNGKey(52), S=100)
    slice_sampler = MultiDimSliceSampler(model=model, num_restrict_dims=model.U_ndims)
    exact_ns = ExactNestedSampler(model=model, num_live_points=200, num_parallel_samplers=1,
                                  max_samples=1e4, refinement_slice_sampler=slice_sampler)
    with Timer():
        print("Using Multi Dimensional Slice Sampling")
        termination_reason, state = exact_ns(random.PRNGKey(42),
                                             term_cond=TerminationCondition(live_evidence_frac=1e-4))
        results = exact_ns.to_results(state, termination_reason)
        exact_ns.summary(results)
        num_slices_multi = results.total_num_slices
    assert jnp.isclose(results.log_Z_mean, true_logZ, atol=1.75 * results.log_Z_uncert)

    slice_sampler = UniDimSliceSampler(model=model, midpoint_shrink=True)
    exact_ns = ExactNestedSampler(model=model, num_live_points=200, num_parallel_samplers=1,
                                  max_samples=1e4, refinement_slice_sampler=slice_sampler)
    with Timer():
        print("Using Uni Dimensional Slice Sampling")
        termination_reason, state = exact_ns(random.PRNGKey(42),
                                             term_cond=TerminationCondition(live_evidence_frac=1e-4))
        results = exact_ns.to_results(state, termination_reason)
        exact_ns.summary(results)
        num_slices_uni = results.total_num_slices
    assert jnp.isclose(results.log_Z_mean, true_logZ, atol=1.75 * results.log_Z_uncert)
    # TODO(Joshuaalbert): broken test, probably means multidim slice sampling is not robust to structure.
    # assert num_slices_multi < num_slices_uni


def test_shrinkage():
    n = 2

    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=1))
        return x

    def log_likelihood(x):
        return jnp.log(1. - x ** n)

    def exact_X(L):
        return (1. - L) ** (1. / n)

    model = Model(prior_model=prior_model,
                  log_likelihood=log_likelihood)
    exact_ns = ExactNestedSampler(model=model, num_live_points=50, num_parallel_samplers=1,
                                  max_samples=1000)

    termination_reason, state = exact_ns(random.PRNGKey(42),
                                         term_cond=TerminationCondition(live_evidence_frac=1e-4))
    results = exact_ns.to_results(state, termination_reason)

    # print(list(results.log_X_mean[:results.total_num_samples]))
    # print(list(results.log_L_samples[:results.total_num_samples]))
    # print(list(jnp.log(exact_X(jnp.exp(results.log_L_samples[:results.total_num_samples])))))
    diff = results.log_X_mean - jnp.log(exact_X(jnp.exp(results.log_L_samples)))
    diff = jnp.where(jnp.isfinite(diff), diff, jnp.nan)
    # print(jnp.nanstd(diff))
    assert jnp.nanstd(diff) < 0.26


def test_nested_sampling_plateau():
    def log_likelihood(x):
        return 0.

    def prior_model() -> PriorModelGen:
        x = yield Prior(tfpd.Uniform(low=0, high=1))
        return x

    model = Model(prior_model=prior_model,
                  log_likelihood=log_likelihood)
    exact_ns = ExactNestedSampler(model=model, num_live_points=50,
                                  max_samples=1000)

    termination_reason, state = exact_ns(random.PRNGKey(42),
                                         term_cond=TerminationCondition(live_evidence_frac=1e-4))
    results = exact_ns.to_results(state, termination_reason)
    exact_ns.summary(results)

    log_Z_samples = evidence_posterior_samples(random.PRNGKey(42),
                                               results.num_live_points_per_sample,
                                               results.log_L_samples, S=1000)

    true_log_Z_uncert = jnp.std(log_Z_samples)
    print(f"true logZ uncert: {true_log_Z_uncert}")

    assert jnp.bitwise_not(jnp.isnan(results.log_Z_mean))
    assert jnp.isclose(results.log_Z_mean, 0., atol=1.75 * true_log_Z_uncert)



from etils.array_types import PRNGKey, FloatArray
from jax import random, numpy as jnp
from jax.lax import while_loop

from jaxns.model import Model
from jaxns.types import Sample, int_type

__all__ = ['UniformSampler']


class UniformSampler:
    def __init__(self, model: Model):
        self.model = model

    def single_sample(self, key: PRNGKey, log_L_constraint: FloatArray) -> Sample:
        """
        Produces a single sample from the joint-prior and computes the likelihood.

        Args:
            key: PRNG key
            log_L_constraint: the log likelihood constraint to sample above

        Returns:
            a sample
        """
        key, sample_key = random.split(key, 2)
        U = self.model.sample_U(key=sample_key)
        log_L = self.model.forward(U=U)
        num_likelihood_evals = jnp.asarray(1, int_type)
        done = (log_L > log_L_constraint)

        def body(state):
            (_, key, _, _, num_likelihood_evals) = state
            key, sample_key = random.split(key, 2)
            U = self.model.sample_U(key=sample_key)
            log_L = self.model.forward(U=U)
            num_likelihood_evals += jnp.asarray(1, int_type)
            done = (log_L > log_L_constraint)
            return (done, key, U, log_L, num_likelihood_evals)

        (_, _, U, log_L, num_likelihood_evals) = while_loop(lambda s: jnp.bitwise_not(s[0]),
                                                            body,
                                                            (done, key, U, log_L, num_likelihood_evals))
        sample = Sample(point_U=U,
                        log_L_constraint=log_L_constraint,
                        log_L=log_L,
                        num_likelihood_evaluations=num_likelihood_evals,
                        num_slices=jnp.asarray(0, int_type),
                        iid=jnp.asarray(True, jnp.bool_))
        return sample
